# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import zipfile
from abc import abstractmethod
from contextlib import contextmanager

from pex.atomic_directory import atomic_directory
from pex.common import (
    PermPreservingZipFile,
    is_script,
    open_zip,
    safe_copy,
    safe_mkdir,
    safe_mkdtemp,
)
from pex.enum import Enum
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.variables import unzip_dir

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, Dict, Iterator, Optional, Text, Tuple, Type

    from pex.pex_info import PexInfo


BOOTSTRAP_DIR = ".bootstrap"
DEPS_DIR = ".deps"
PEX_INFO_PATH = "PEX-INFO"
PEX_LAYOUT_PATH = "PEX-LAYOUT"


class Layout(Enum["Layout.Value"]):
    class Value(Enum.Value):
        @classmethod
        def try_load(cls, pex_directory):
            # type: (str) -> Optional[Layout.Value]
            layout = os.path.join(pex_directory, PEX_LAYOUT_PATH)
            if not os.path.isfile(layout):
                return None
            with open(layout) as fp:
                return Layout.for_value(fp.read().strip())

        def record(self, pex_directory):
            # type: (str) -> None
            with open(os.path.join(pex_directory, PEX_LAYOUT_PATH), "w") as fp:
                fp.write(self.value)

    ZIPAPP = Value("zipapp")
    PACKED = Value("packed")
    LOOSE = Value("loose")

    @classmethod
    def identify(cls, pex):
        # type: (str) -> Layout.Value
        """Assumes pex is a valid PEX and identifies its layout."""
        if zipfile.is_zipfile(pex) and is_script(
            pex,
            # N.B.: A PEX file need not be executable since it can always be run via `python a.pex`.
            check_executable=False,
        ):
            return cls.ZIPAPP

        if os.path.isdir(pex) and zipfile.is_zipfile(os.path.join(pex, BOOTSTRAP_DIR)):
            return cls.PACKED

        return cls.LOOSE

    @classmethod
    def identify_original(cls, pex):
        # type: (str) -> Layout.Value
        layout = cls.identify(pex)
        if layout is not Layout.LOOSE:
            return layout
        return cls.Value.try_load(pex) or Layout.LOOSE


class _Layout(object):
    def __init__(
        self,
        layout,  # type: Layout.Value
        path,  # type: str
    ):
        # type: (...) -> None
        self._layout = layout
        self._path = os.path.normpath(path)

    @property
    def type(self):
        # type: () -> Layout.Value
        return self._layout

    @property
    def path(self):
        # type: () -> str
        return self._path

    def bootstrap_strip_prefix(self):
        # type: () -> Optional[str]
        return None

    @abstractmethod
    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    def dist_strip_prefix(self, dist_name):
        # type: (str) -> Optional[str]
        return None

    @abstractmethod
    def dist_size(
        self,
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> int
        raise NotImplementedError()

    @abstractmethod
    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> None
        raise NotImplementedError()

    @abstractmethod
    def wheel_file_path(self, dist_relpath):
        # type: (str) -> str
        raise NotImplementedError()

    @abstractmethod
    def extract_code(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def extract_main(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    def record(self, dest_dir):
        # type: (str) -> None
        self._layout.record(dest_dir)


def _install_distribution(
    distribution_info,  # type: Tuple[str, str]
    layout,  # type: _Layout
    pex_info,  # type: PexInfo
    work_dir,  # type: str
    install_to,  # type: str
):
    # type: (...) -> str

    location, sha = distribution_info
    is_wheel_file = pex_info.deps_are_wheel_files
    spread_dest = os.path.join(pex_info.install_cache, sha, location)
    dist_relpath = os.path.join(DEPS_DIR, location)
    source = None if is_wheel_file else layout.dist_strip_prefix(location)
    symlink_src = os.path.relpath(
        spread_dest,
        os.path.join(install_to, os.path.dirname(dist_relpath)),
    )
    symlink_dest = os.path.join(work_dir, dist_relpath)

    with atomic_directory(spread_dest, source=source) as spread_chroot:
        if not spread_chroot.is_finalized():
            layout.extract_dist(
                dest_dir=spread_chroot.work_dir,
                dist_relpath=dist_relpath,
                is_wheel_file=is_wheel_file,
            )

    safe_mkdir(os.path.dirname(symlink_dest))
    os.symlink(symlink_src, symlink_dest)
    return location


def _ensure_distributions_installed_serial(
    layout,  # type: _Layout
    pex_info,  # type: PexInfo
    work_dir,  # type: str
    install_to,  # type: str
):
    # type: (...) -> None

    for item in pex_info.distributions.items():
        with TRACER.timed("Installing {wheel}".format(wheel=item[0]), V=2):
            _install_distribution(item, layout, pex_info, work_dir, install_to)


def _ensure_distributions_installed_parallel(
    layout,  # type: _Layout
    pex_info,  # type: PexInfo
    work_dir,  # type: str
    install_to,  # type: str
    max_jobs,  # type: int
):
    # type: (...) -> None

    import functools

    from pex.jobs import map_parallel

    install_distribution = functools.partial(
        _install_distribution,
        layout=layout,
        pex_info=pex_info,
        work_dir=work_dir,
        install_to=install_to,
    )

    map_parallel(
        inputs=pex_info.distributions.items(),
        function=install_distribution,
        noun="wheel",
        verb="install",
        verb_past="installed",
        max_jobs=max_jobs,
        costing_function=lambda item: layout.dist_size(
            os.path.join(DEPS_DIR, item[0]), is_wheel_file=pex_info.deps_are_wheel_files
        ),
    )


# This value was found via experiment on a single laptop with 16 cores and SSD storage. The
# threshold that needs to be overcome is the startup and communication overhead of a Python
# multiprocessing pool. It's completely unclear this is a good value in general let alone the
# heuristic using it is reasonable; as such the user-facing documentation makes it clear the
# heuristic can be tweaked from release to release in order to try to improve performance.
AVERAGE_DISTRIBUTION_SIZE_PARALLEL_JOB_THRESHOLD = 1 * 1024 * 1024  # ~1MB


def _ensure_distributions_installed(
    layout,  # type: _Layout
    pex_info,  # type: PexInfo
    work_dir,  # type: str
    install_to,  # type: str
):
    # type: (...) -> None

    dist_count = len(pex_info.distributions)
    if dist_count == 0:
        return

    install_serial = dist_count == 1 or pex_info.max_install_jobs == 1
    if not install_serial and pex_info.max_install_jobs == -1:
        total_size = sum(
            layout.dist_size(os.path.join(DEPS_DIR, location), pex_info.deps_are_wheel_files)
            for location in pex_info.distributions
        )
        average_distribution_size = total_size // dist_count
        install_serial = (
            average_distribution_size < AVERAGE_DISTRIBUTION_SIZE_PARALLEL_JOB_THRESHOLD
        )
        if install_serial:
            TRACER.log(
                "Installing {count} distributions in serial based on average distribution "
                "size of {avg_size} bytes".format(
                    count=dist_count, avg_size=average_distribution_size
                )
            )
        else:
            TRACER.log(
                "Installing {count} distributions in parallel based on average distribution "
                "size of {avg_size} bytes".format(
                    count=dist_count, avg_size=average_distribution_size
                )
            )

    if install_serial:
        _ensure_distributions_installed_serial(
            layout=layout, pex_info=pex_info, work_dir=work_dir, install_to=install_to
        )
    else:
        max_jobs = 0 if pex_info.max_install_jobs == -1 else pex_info.max_install_jobs
        _ensure_distributions_installed_parallel(
            layout=layout,
            pex_info=pex_info,
            work_dir=work_dir,
            install_to=install_to,
            max_jobs=max_jobs,
        )


def _ensure_installed(
    layout,  # type: _Layout
    pex_root,  # type: str
    pex_hash,  # type: str
):
    # type: (...) -> str
    if layout.type is Layout.LOOSE:
        from pex.pex_info import PexInfo

        pex_info = PexInfo.from_pex(layout.path)
        if not pex_info.distributions or not pex_info.deps_are_wheel_files:
            # A loose PEX with no dependencies or dependencies that are pre-installed wheel chroots
            # is in the canonical form of a PEX executable zipapp already and needs no install.
            return layout.path

    with TRACER.timed("Laying out {}".format(layout)):
        pex = layout.path
        install_to = unzip_dir(pex_root=pex_root, pex_hash=pex_hash)
        with atomic_directory(install_to) as chroot:
            if not chroot.is_finalized():
                from pex.variables import ENV

                with ENV.patch(PEX_ROOT=pex_root), TRACER.timed(
                    "Installing {} to {}".format(pex, install_to)
                ):
                    from pex.pex_info import PexInfo

                    pex_info = PexInfo.from_pex(pex)
                    pex_info.update(PexInfo.from_env())

                    bootstrap_cache = pex_info.bootstrap_cache
                    if bootstrap_cache is None:
                        raise AssertionError(
                            "Expected bootstrap_cache to be populated for {}.".format(layout)
                        )
                    code_hash = pex_info.code_hash
                    if code_hash is None:
                        raise AssertionError(
                            "Expected code_hash to be populated for {}.".format(layout)
                        )

                    with atomic_directory(
                        bootstrap_cache, source=layout.bootstrap_strip_prefix()
                    ) as bootstrap_zip_chroot:
                        if not bootstrap_zip_chroot.is_finalized():
                            layout.extract_bootstrap(bootstrap_zip_chroot.work_dir)
                    os.symlink(
                        os.path.join(os.path.relpath(bootstrap_cache, install_to)),
                        os.path.join(chroot.work_dir, BOOTSTRAP_DIR),
                    )

                    _ensure_distributions_installed(
                        layout=layout,
                        pex_info=pex_info,
                        work_dir=chroot.work_dir,
                        install_to=install_to,
                    )

                    code_dest = os.path.join(pex_info.zip_unsafe_cache, code_hash)
                    with atomic_directory(code_dest) as code_chroot:
                        if not code_chroot.is_finalized():
                            layout.extract_code(code_chroot.work_dir)
                    for path in os.listdir(code_dest):
                        os.symlink(
                            os.path.join(os.path.relpath(code_dest, install_to), path),
                            os.path.join(chroot.work_dir, path),
                        )

                    layout.extract_pex_info(chroot.work_dir)
                    layout.extract_main(chroot.work_dir)
                    layout.record(chroot.work_dir)
        return install_to


class _ZipAppPEX(_Layout):
    def __init__(
        self,
        path,  # type: str
    ):
        # type: (...) -> None
        super(_ZipAppPEX, self).__init__(Layout.ZIPAPP, path)
        self._zfp = None  # type: Optional[zipfile.ZipFile]
        self._names = None  # type: Optional[Tuple[Text, ...]]

    def close(self):
        # type: () -> None
        if self._zfp is not None:
            self._zfp.close()

    def open(self):
        # type: () -> zipfile.ZipFile
        if self._zfp is None:
            self._zfp = PermPreservingZipFile(self.path)
        return self._zfp

    def __getstate__(self):
        # type: () -> Dict[str, Any]
        state = self.__dict__.copy()
        # We can't pickle an open file handle; so don't try.
        state["_zfp"] = None
        return state

    def __enter__(self):
        # type: () -> _ZipAppPEX
        self.open()
        return self

    def __exit__(
        self,
        _exc_type,  # type: Optional[Type]
        _exc_val,  # type: Optional[Any]
        _exc_tb,  # type: Optional[TracebackType]
    ):
        # type: (...) -> None
        self.close()

    def __del__(self):
        # type: () -> None
        self.close()

    @property
    def zfp(self):
        # type: () -> zipfile.ZipFile
        return self.open()

    @property
    def names(self):
        # type: () -> Tuple[Text, ...]
        if self._names is None:
            self._names = tuple(self.zfp.namelist())
        return self._names

    def bootstrap_strip_prefix(self):
        # type: () -> Optional[str]
        return BOOTSTRAP_DIR

    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        for name in self.names:
            if name.startswith(BOOTSTRAP_DIR) and not name.endswith("/"):
                self.zfp.extract(name, dest_dir)

    def dist_strip_prefix(self, dist_name):
        # type: (str) -> Optional[str]
        return os.path.join(DEPS_DIR, dist_name)

    def dist_size(
        self,
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> int
        if is_wheel_file:
            return self.zfp.getinfo(dist_relpath).file_size
        else:
            return sum(
                self.zfp.getinfo(name).file_size
                for name in self.names
                if name.startswith(dist_relpath)
            )

    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> None
        if is_wheel_file:
            from pex.pep_427 import install_wheel_chroot

            install_wheel_chroot(self.wheel_file_path(dist_relpath), dest_dir)
        else:
            for name in self.names:
                if name.startswith(dist_relpath) and not name.endswith("/"):
                    self.zfp.extract(name, dest_dir)

    def wheel_file_path(self, dist_relpath):
        # type: (str) -> str
        extract_chroot = safe_mkdtemp()
        self.zfp.extract(dist_relpath, extract_chroot)
        return os.path.join(extract_chroot, dist_relpath)

    def extract_code(self, dest_dir):
        # type: (str) -> None
        for name in self.names:
            if name not in ("__main__.py", PEX_INFO_PATH) and not name.startswith(
                (BOOTSTRAP_DIR, DEPS_DIR)
            ):
                self.zfp.extract(name, dest_dir)

    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        self.zfp.extract(PEX_INFO_PATH, dest_dir)

    def extract_main(self, dest_dir):
        # type: (str) -> None
        self.zfp.extract("__main__.py", dest_dir)

    def __str__(self):
        return "PEX zipfile {}".format(self.path)


class _PackedPEX(_Layout):
    def __init__(self, path):
        # type: (str) -> None
        super(_PackedPEX, self).__init__(Layout.PACKED, path)

    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        with open_zip(os.path.join(self._path, BOOTSTRAP_DIR)) as zfp:
            zfp.extractall(dest_dir)

    def dist_size(
        self,
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> int
        return os.path.getsize(os.path.join(self._path, dist_relpath))

    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        # type: (...) -> None
        dist_path = self.wheel_file_path(dist_relpath)
        if is_wheel_file:
            from pex.pep_427 import install_wheel_chroot

            with TRACER.timed("Installing wheel file {}".format(dist_relpath)):
                install_wheel_chroot(dist_path, dest_dir)
        else:
            with TRACER.timed("Installing zipped wheel install {}".format(dist_relpath)):
                with open_zip(dist_path) as zfp:
                    zfp.extractall(dest_dir)

    def wheel_file_path(self, dist_relpath):
        # type: (str) -> str
        return os.path.join(self._path, dist_relpath)

    def extract_code(self, dest_dir):
        # type: (str) -> None
        for root, dirs, files in os.walk(self._path):
            rel_root = os.path.relpath(root, self._path)
            if root == self._path:
                dirs[:] = [d for d in dirs if d != DEPS_DIR]
                files[:] = [
                    f for f in files if f not in ("__main__.py", PEX_INFO_PATH, BOOTSTRAP_DIR)
                ]
            for d in dirs:
                safe_mkdir(os.path.join(dest_dir, rel_root, d))
            for f in files:
                safe_copy(
                    os.path.join(root, f),
                    os.path.join(dest_dir, rel_root, f),
                )

    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, PEX_INFO_PATH), os.path.join(dest_dir, PEX_INFO_PATH))

    def extract_main(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, "__main__.py"), os.path.join(dest_dir, "__main__.py"))

    def __str__(self):
        return "Spread PEX directory {}".format(self._path)


class _LoosePEX(_Layout):
    def __init__(self, path):
        super(_LoosePEX, self).__init__(Layout.LOOSE, path)

    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        bootstrap_dir = os.path.join(self._path, BOOTSTRAP_DIR)
        for root, dirs, files in os.walk(bootstrap_dir):
            rel_root = os.path.relpath(root, bootstrap_dir)
            for d in dirs:
                safe_mkdir(os.path.join(dest_dir, rel_root, d))
            for f in files:
                safe_copy(os.path.join(root, f), os.path.join(dest_dir, rel_root, f))

    def dist_size(
        self,
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        assert (
            is_wheel_file
        ), "Expected loose layout install to be skipped when deps are pre-installed wheel chroots."
        return os.path.getsize(os.path.join(self._path, dist_relpath))

    def extract_dist(
        self,
        dest_dir,
        dist_relpath,  # type: str
        is_wheel_file,  # type: bool
    ):
        assert (
            is_wheel_file
        ), "Expected loose layout install to be skipped when deps are pre-installed wheel chroots."
        from pex.pep_427 import install_wheel_chroot

        with TRACER.timed("Installing wheel file {}".format(dist_relpath)):
            install_wheel_chroot(self.wheel_file_path(dist_relpath), dest_dir)

    def wheel_file_path(self, dist_relpath):
        # type: (str) -> str
        return os.path.join(self._path, dist_relpath)

    def extract_code(self, dest_dir):
        # type: (str) -> None
        for root, dirs, files in os.walk(self._path):
            rel_root = os.path.relpath(root, self._path)
            if root == self._path:
                dirs[:] = [d for d in dirs if d not in (DEPS_DIR, BOOTSTRAP_DIR)]
                files[:] = [f for f in files if f not in ("__main__.py", PEX_INFO_PATH)]
            for d in dirs:
                safe_mkdir(os.path.join(dest_dir, rel_root, d))
            for f in files:
                safe_copy(
                    os.path.join(root, f),
                    os.path.join(dest_dir, rel_root, f),
                )

    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, PEX_INFO_PATH), os.path.join(dest_dir, PEX_INFO_PATH))

    def extract_main(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, "__main__.py"), os.path.join(dest_dir, "__main__.py"))

    def __str__(self):
        return "Loose PEX directory {}".format(self._path)


@contextmanager
def identify_layout(pex):
    # type: (str) -> Iterator[_Layout]

    layout = Layout.identify(pex)
    if Layout.ZIPAPP is layout:
        with _ZipAppPEX(pex) as zip_app_pex:
            yield zip_app_pex
    elif Layout.PACKED is layout:
        yield _PackedPEX(pex)
    elif Layout.LOOSE is layout:
        yield _LoosePEX(pex)
    else:
        raise AssertionError("Un-handled PEX layout type: {layout}".format(layout=layout))


def ensure_installed(
    pex,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
):
    # type: (...) -> str
    """Installs a zipapp or packed PEX into the pex root as a loose PEX.

    Returns the path of the installed PEX or `None` if the PEX needed no installation and can be
    executed directly.
    """
    with identify_layout(pex) as layout:
        return _ensure_installed(layout, pex_root, pex_hash)



# Copyright 2014 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

import pytest

from pex.dist_metadata import CallableEntryPoint, Distribution, EntryPoint
from pex.finders import (
    DistributionScript,
    get_entry_point_from_console_script,
    get_script_from_distributions,
)
from pex.pep_376 import InstalledWheel
from pex.pep_427 import install_wheel_chroot
from pex.typing import TYPE_CHECKING
from testing.dist_metadata import create_dist_metadata

if TYPE_CHECKING:
    from typing import Any, Dict, Text, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


# In-part, tests a bug where the wheel distribution name has dashes as reported in:
#   https://github.com/pex-tool/pex/issues/443
#   https://github.com/pex-tool/pex/issues/551
def test_get_script_from_distributions(tmpdir):
    # type: (Any) -> None

    def assert_script(dist):
        # type: (Distribution) -> Tuple[Distribution, DistributionScript]
        assert "aws-cfn-bootstrap" == dist.project_name

        dist_script = get_script_from_distributions("cfn-signal", [dist])
        assert dist_script is not None
        assert dist_script.dist is dist
        assert dist_script.read_contents().startswith(
            b"#!"
        ), "Expected a `scripts`-style script w/shebang."

        assert None is get_script_from_distributions("non_existent_script", [dist])
        return dist, dist_script

    whl_path = "./tests/example_packages/aws_cfn_bootstrap-1.4-py2-none-any.whl"
    _, dist_script = assert_script(Distribution.load(whl_path))
    assert "aws_cfn_bootstrap-1.4.data/scripts/cfn-signal" == dist_script.path

    install_dir = os.path.join(str(tmpdir), os.path.basename(whl_path))
    install_wheel_chroot(wheel_path=whl_path, destination=install_dir)
    installed_wheel_dist, dist_script = assert_script(Distribution.load(install_dir))
    assert InstalledWheel.load(install_dir).stashed_path("bin/cfn-signal") == dist_script.path


def create_dist(
    key,  # str
    console_script_entry,  # type: str
):
    # type: (...) -> Distribution
    entry_point = EntryPoint.parse(console_script_entry)

    @attr.s(frozen=True)
    class FakeDist(Distribution):
        def get_entry_map(self):
            # type: () -> Dict[Text, Dict[Text, EntryPoint]]
            return {"console_scripts": {entry_point.name: entry_point}}

    location = os.getcwd()
    return FakeDist(
        location=location,
        metadata=create_dist_metadata(key, "1.0", location=location),
    )


def test_get_entry_point_from_console_script():
    # type: () -> None
    dists = [
        create_dist(key="fake", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]

    dist_entrypoint = get_entry_point_from_console_script("bob", dists)
    assert dist_entrypoint is not None
    assert (
        CallableEntryPoint(name="bob", module="bob.main", attrs=("run",))
        == dist_entrypoint.entry_point
    )
    assert dist_entrypoint.dist in dists


def test_get_entry_point_from_console_script_conflict():
    # type: () -> None
    dists = [
        create_dist(key="bob", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]
    with pytest.raises(RuntimeError):
        get_entry_point_from_console_script("bob", dists)


def test_get_entry_point_from_console_script_dne():
    # type: () -> None
    dists = [
        create_dist(key="bob", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]
    assert get_entry_point_from_console_script("jane", dists) is None



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.requirements import ArchiveScheme
from pex.resolve.resolved_requirement import ArtifactURL, Fingerprint


def test_artifact_url_escaping():
    # type: () -> None

    with_space = ArtifactURL.parse("file:///a/path/with%20space")
    assert "file" == with_space.scheme
    assert "/a/path/with space" == with_space.path
    assert not with_space.is_wheel
    assert "file:///a/path/with%20space" == with_space.raw_url
    assert "file:///a/path/with%20space" == with_space.download_url
    assert "file:///a/path/with space" == with_space.normalized_url
    assert not with_space.fragment_parameters
    assert not with_space.fingerprints
    assert not with_space.fingerprint


def test_artifact_url_with_hash():
    # type: () -> None
    with_hash = ArtifactURL.parse(
        "https://example.com/a/path/with%2Bplus.whl"
        "?query=how"
        "#sha1=abcd1234&foo=bar&sha256=1234abcd&foo=baz"
    )
    assert ArchiveScheme.HTTPS is with_hash.scheme
    assert "/a/path/with+plus.whl" == with_hash.path
    assert with_hash.is_wheel
    assert (
        "https://example.com/a/path/with%2Bplus.whl"
        "?query=how"
        "#sha1=abcd1234&foo=bar&sha256=1234abcd&foo=baz" == with_hash.raw_url
    )
    assert (
        "https://example.com/a/path/with%2Bplus.whl?query=how#foo=bar&foo=baz"
        == with_hash.download_url
    )
    assert "https://example.com/a/path/with+plus.whl" == with_hash.normalized_url
    assert {"foo": ["bar", "baz"]} == {
        name: list(values) for name, values in with_hash.fragment_parameters.items()
    }
    assert (
        Fingerprint("sha256", "1234abcd"),
        Fingerprint("sha1", "abcd1234"),
    ) == with_hash.fingerprints
    assert Fingerprint("sha256", "1234abcd") == with_hash.fingerprint



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import subprocess
from textwrap import dedent

from colors import colors

from pex.common import safe_open
from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any


def test_pex_with_editable(tmpdir):
    # type: (Any) -> None

    project_dir = os.path.join(str(tmpdir), "project")
    with safe_open(os.path.join(project_dir, "example.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import sys

                import colors


                def colorize(*messages):
                    return colors.green(" ".join(messages))


                if __name__ == "__main__":
                    print(colorize(*sys.argv[1:]))
                    sys.exit(0)
                """
            )
        )
    with safe_open(os.path.join(project_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from setuptools import setup


                setup(
                    name="example",
                    version="0.1.0",
                    py_modules=["example"],
                )
                """
            )
        )

    requirements = os.path.join(project_dir, "requirements.txt")
    with safe_open(requirements, "w") as fp:
        fp.write(
            dedent(
                """\
                ansicolors==1.1.8
                -e file://{project_dir}
                """
            ).format(project_dir=project_dir)
        )

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["-r", requirements, "-m", "example", "-o", pex]).assert_success()
    output = (
        subprocess.check_output(args=[pex, "A", "wet", "duck", "flies", "at", "night!"])
        .decode("utf-8")
        .strip()
    )
    assert colors.green("A wet duck flies at night!") == output, output



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.cli.command import BuildTimeCommand
from pex.cli.commands.docs import Docs
from pex.cli.commands.interpreter import Interpreter
from pex.cli.commands.lock import Lock
from pex.cli.commands.venv import Venv
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Type


def all_commands():
    # type: () -> Iterable[Type[BuildTimeCommand]]
    return Docs, Interpreter, Lock, Venv



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import glob
import os
import subprocess

import pytest

from pex.atomic_directory import atomic_directory
from pex.common import temporary_dir
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv
from testing import PY310, data, ensure_python_interpreter, make_env, run_pex_command
from testing.mitmproxy import Proxy

if TYPE_CHECKING:
    from typing import Any, Callable, Iterator


@pytest.fixture(scope="session")
def pexpect_timeout():
    # type: () -> int

    # The default here of 5 provides enough margin for PyPy which has slow startup.
    return int(os.environ.get("_PEX_PEXPECT_TIMEOUT", "5"))


@pytest.fixture(scope="session")
def is_pytest_xdist(worker_id):
    # type: (str) -> bool
    return worker_id != "master"


@pytest.fixture(scope="session")
def shared_integration_test_tmpdir(
    tmpdir_factory,  # type: Any
    is_pytest_xdist,  # type: bool
):
    # type: (...) -> str
    tmpdir = str(tmpdir_factory.getbasetemp())

    # We know pytest-xdist creates a subdir under the pytest session tmp dir for each worker; so we
    # go up a level to lock a directory all workers can use.
    if is_pytest_xdist:
        tmpdir = os.path.dirname(tmpdir)

    return os.path.join(tmpdir, "shared_integration_test_tmpdir")


@pytest.fixture(scope="session")
def pex_bdist(
    pex_project_dir,  # type: str
    shared_integration_test_tmpdir,  # type: str
):
    # type: (...) -> str
    pex_bdist_chroot = os.path.join(shared_integration_test_tmpdir, "pex_bdist_chroot")
    wheels_dir = os.path.join(pex_bdist_chroot, "wheels_dir")
    with atomic_directory(pex_bdist_chroot) as chroot:
        if not chroot.is_finalized():
            pex_pex = os.path.join(chroot.work_dir, "pex.pex")
            run_pex_command(
                args=[pex_project_dir, "-o", pex_pex, "--include-tools"]
            ).assert_success()
            extract_dir = os.path.join(chroot.work_dir, "wheels_dir")
            subprocess.check_call(
                args=[pex_pex, "repository", "extract", "-f", extract_dir],
                env=make_env(PEX_TOOLS=True),
            )
    wheels = glob.glob(os.path.join(wheels_dir, "pex-*.whl"))
    assert 1 == len(wheels)
    return wheels[0]


@pytest.fixture
def tmp_workdir():
    # type: () -> Iterator[str]
    cwd = os.getcwd()
    with temporary_dir() as tmpdir:
        os.chdir(tmpdir)
        try:
            yield os.path.realpath(tmpdir)
        finally:
            os.chdir(cwd)


@pytest.fixture(scope="session")
def mitmdump_venv(shared_integration_test_tmpdir):
    # type: (str) -> Virtualenv
    mitmproxy_venv_dir = os.path.join(shared_integration_test_tmpdir, "mitmproxy")
    with atomic_directory(mitmproxy_venv_dir) as atomic_venvdir:
        if not atomic_venvdir.is_finalized():
            python = ensure_python_interpreter(PY310)
            Virtualenv.create_atomic(
                venv_dir=atomic_venvdir,
                interpreter=PythonInterpreter.from_binary(python),
                force=True,
            )
            mitmproxy_lock = data.path("locks", "mitmproxy.lock.json")
            subprocess.check_call(
                args=[
                    python,
                    "-m",
                    "pex.cli",
                    "venv",
                    "create",
                    "-d",
                    atomic_venvdir.work_dir,
                    "--lock",
                    mitmproxy_lock,
                ]
            )
    return Virtualenv(mitmproxy_venv_dir)


@pytest.fixture
def proxy(tmpdir):
    # type: (Any) -> Proxy
    config_dir = os.path.join(str(tmpdir), "mitmdump-cfg")
    os.mkdir(config_dir)
    return Proxy.configured(config_dir=config_dir)


@pytest.fixture
def clone(tmpdir):
    # type: (Any) -> Callable[[str, str], str]

    def _clone(
        git_project_url,  # type: str
        commit,  # type: str
    ):
        project_dir = os.path.join(str(tmpdir), "project")

        subprocess.check_call(args=["git", "clone", git_project_url, project_dir])
        subprocess.check_call(
            args=["git", "config", "advice.detachedHead", "false"], cwd=project_dir
        )
        subprocess.check_call(args=["git", "checkout", commit], cwd=project_dir)
        return project_dir

    return _clone



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict


def patch():
    # type: () -> None

    from pip._vendor.packaging import markers  # type: ignore[import]

    from pex.exceptions import production_assert
    from pex.pip.foreign_platform import EvaluationEnvironment, PatchContext

    evaluation_environment = PatchContext.load_evaluation_environment()

    def _get_env(
        environment,  # type: Dict[Any, Any]
        name,  # type: Any
    ):
        # type: (...) -> Any
        production_assert(
            isinstance(environment, EvaluationEnvironment),
            "Expected environment to come from the {function} function, "
            "which we patch to return {expected_type}, but was {actual_type}".format(
                function=markers.default_environment,
                expected_type=EvaluationEnvironment,
                actual_type=type(environment),
            ),
        )
        return environment[name]

    # Works with all Pip vendored packaging distributions.
    markers.default_environment = evaluation_environment.default
    # Covers Pip<24.1 vendored packaging.
    markers._get_env = _get_env

    original_eval_op = markers._eval_op

    def _eval_op(
        lhs,  # type: Any
        op,  # type: Any
        rhs,  # type: Any
    ):
        # type: (...) -> Any
        evaluation_environment.raise_if_missing(lhs)
        evaluation_environment.raise_if_missing(rhs)
        return original_eval_op(lhs, op, rhs)

    markers._eval_op = _eval_op



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import platform
import sys
from textwrap import dedent

from pex.version import __version__

_ASSERT_DETAILS = (
    dedent(
        """\
        Pex {version}
        platform: {platform}
        python: {python_version}
        argv: {argv}
        """
    )
    .format(
        version=__version__, platform=platform.platform(), python_version=sys.version, argv=sys.argv
    )
    .strip()
)

_ASSERT_ADVICE = dedent(
    """\
    The error reported above resulted from an unexpected error which you should
    never encounter.

    Firstly, please accept our apology!

    If you could file an issue with the error and details above, we'd be
    grateful. You can do that at https://github.com/pex-tool/pex/issues/new and
    redact or amend any details that expose sensitive information.
    """
).strip()


def reportable_unexpected_error_msg(msg=""):
    # type: (str) -> str

    message = [msg, "---", _ASSERT_DETAILS]
    pex = os.environ.get("PEX")
    if pex:
        try:
            import json

            from pex.pex_info import PexInfo

            pex_info = PexInfo.from_pex(pex)
            pex_info.update(PexInfo.from_env())
            pex_info_json = json.dumps(pex_info.as_json_dict(), indent=2)
        except Exception:
            pass
        else:
            message.append("PEX-INFO:")
            message.append(pex_info_json)
    message.append("---")
    message.append(_ASSERT_ADVICE)

    return "\n".join(message)


def production_assert(
    condition,  # type: bool
    msg="",  # type: str
):
    # type: (...) -> None

    if not condition:
        raise AssertionError(reportable_unexpected_error_msg(msg=msg))



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import re

import pytest

from pex.atomic_directory import atomic_directory
from pex.pep_427 import InstallableType
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.locked_resolve import FileArtifact
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolved_requirement import Pin
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3
from testing.find_links import FindLinksRepo

if TYPE_CHECKING:
    from typing import Any, List


@pytest.fixture(scope="session")
def find_links(shared_integration_test_tmpdir):
    # type: (str) -> str

    test_issue_2343_chroot = os.path.join(shared_integration_test_tmpdir, "test_issue_2343_chroot")
    with atomic_directory(test_issue_2343_chroot) as chroot:
        if not chroot.is_finalized():
            pip_version = PipVersion.DEFAULT
            find_links = os.path.join(chroot.work_dir, "find_links")
            find_links_repo = FindLinksRepo.create(find_links, pip_version)

            # N.B.: Since we are setting up a find links repo for offline lock resolves, we grab one
            # distribution online to allow the current Pip version to bootstrap itself if needed.
            result = find_links_repo.resolver.resolve_requirements(
                [
                    "ansicolors==1.1.8",
                    pip_version.setuptools_requirement,
                    pip_version.wheel_requirement,
                ],
                result_type=InstallableType.WHEEL_FILE,
            )
            for resolved_distribution in result.distributions:
                find_links_repo.host(resolved_distribution.distribution.location)

            for version in "1", "2":
                find_links_repo.make_sdist("only_sdist", version=version)
                find_links_repo.make_wheel("both", version=version)
                find_links_repo.make_sdist("both", version=version)

    return os.path.join(test_issue_2343_chroot, "find_links")


@pytest.fixture
def repo_args(find_links):
    # type: (str) -> List[str]
    return [
        "--no-pypi",
        "-f",
        find_links,
    ]


def test_no_build_no_wheel_honored_pex(repo_args):
    # type: (List[str]) -> None

    run_pex_command(
        args=repo_args
        + [
            "--no-build",
            "ansicolors",
            "both",
            "--",
            "-c",
            "import both, colors",
        ]
    ).assert_success()
    run_pex_command(args=repo_args + ["--no-wheel", "ansicolors", "both"]).assert_failure(
        expected_error_re=r".*\bERROR: No matching distribution found for ansicolors\b.*",
        re_flags=re.DOTALL,
    )

    run_pex_command(
        args=repo_args
        + [
            "--no-wheel",
            "only_sdist",
            "both",
            "--",
            "-c",
            "import both, only_sdist",
        ]
    ).assert_success()
    run_pex_command(args=repo_args + ["--no-build", "only_sdist", "both"]).assert_failure(
        expected_error_re=r".*\bERROR: No matching distribution found for only_sdist\b.*",
        re_flags=re.DOTALL,
    )


def test_only_build_honored_pex(repo_args):
    # type: (List[str]) -> None

    run_pex_command(
        args=repo_args
        + [
            "--only-build",
            "sdist_only",
            "--only-build",
            "both",
            "ansicolors",
            "both",
            "only_sdist",
            "--",
            "-c",
            "import both, colors, only_sdist",
        ]
    ).assert_success()
    run_pex_command(args=repo_args + ["--only-build", "ansicolors", "ansicolors"]).assert_failure(
        expected_error_re=r".*\bERROR: No matching distribution found for ansicolors\b.*",
        re_flags=re.DOTALL,
    )


def test_only_wheel_honored_pex(repo_args):
    # type: (List[str]) -> None

    run_pex_command(
        args=repo_args
        + [
            "--only-wheel",
            "ansicolors",
            "--only-wheel",
            "both",
            "ansicolors",
            "both",
            "only_sdist",
            "--",
            "-c",
            "import both, colors, only_sdist",
        ]
    ).assert_success()
    run_pex_command(args=repo_args + ["--only-wheel", "only_sdist", "only_sdist"]).assert_failure(
        expected_error_re=r".*\bERROR: No matching distribution found for only_sdist\b.*",
        re_flags=re.DOTALL,
    )


def assert_lock_single(
    lock,  # type: str
    expected_project_name,  # type: str
    expected_version,  # type: str
    is_source,  # type: bool
):
    # type: (...) -> Lockfile
    lock_file = json_codec.load(lock)
    locked_requirements = [
        locked_requirement
        for locked_resolve in lock_file.locked_resolves
        for locked_requirement in locked_resolve.locked_requirements
    ]

    assert 1 == len(locked_requirements)
    locked_requirement = locked_requirements[0]
    assert (
        Pin(ProjectName(expected_project_name), Version(expected_version)) == locked_requirement.pin
    )

    artifacts = list(locked_requirement.iter_artifacts())
    assert 1 == len(artifacts)
    assert isinstance(artifacts[0], FileArtifact)
    assert is_source == artifacts[0].is_source
    return lock_file


def test_only_build_honored_lock(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        *(
            [
                "lock",
                "create",
            ]
            + repo_args
            + [
                "--style",
                "universal",
                "--only-build",
                "both",
                "both<2",
                "--indent",
                "2",
                "-o",
                lock,
            ]
        )
    ).assert_success()
    lock_file = assert_lock_single(lock, "both", "1", is_source=True)
    assert SortedTuple([ProjectName("both")]) == lock_file.only_builds
    assert SortedTuple() == lock_file.only_wheels

    run_pex3(*(["lock", "update"] + repo_args + ["-R", "both", "--indent", "2", lock]))
    lock_file = assert_lock_single(lock, "both", "2", is_source=True)
    assert SortedTuple([ProjectName("both")]) == lock_file.only_builds
    assert SortedTuple() == lock_file.only_wheels


def test_only_wheel_honored_lock(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        *(
            [
                "lock",
                "create",
            ]
            + repo_args
            + [
                "--style",
                "universal",
                "--only-wheel",
                "both",
                "both>1",
                "--indent",
                "2",
                "-o",
                lock,
            ]
        )
    ).assert_success()
    lock_file = assert_lock_single(lock, "both", "2", is_source=False)
    assert SortedTuple() == lock_file.only_builds
    assert SortedTuple([ProjectName("both")]) == lock_file.only_wheels

    run_pex3(*(["lock", "update"] + repo_args + ["-R", "both<2", "--indent", "2", lock]))
    lock_file = assert_lock_single(lock, "both", "1", is_source=False)
    assert SortedTuple() == lock_file.only_builds
    assert SortedTuple([ProjectName("both")]) == lock_file.only_wheels



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import re
from argparse import ArgumentParser

import pytest

from pex.pex_warnings import PEXWarning
from pex.pip.version import PipVersion
from pex.resolve import resolver_configuration, resolver_options
from pex.resolve.resolver_configuration import (
    BuildConfiguration,
    PexRepositoryConfiguration,
    PipConfiguration,
    ReposConfiguration,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List

    from pex.resolve.resolver_options import ResolverConfiguration


def compute_resolver_configuration(
    parser,  # type: ArgumentParser
    args,  # type: List[str]
):
    # type: (...) -> ResolverConfiguration
    options = parser.parse_args(args=args)
    return resolver_options.configure(options)


def compute_pip_configuration(
    parser,  # type: ArgumentParser
    args,  # type: List[str]
):
    # type: (...) -> PipConfiguration
    resolver_configuration = compute_resolver_configuration(parser, args)
    assert isinstance(resolver_configuration, PipConfiguration)
    return resolver_configuration


def compute_build_configuration(
    parser,  # type: ArgumentParser
    args,  # type: List[str]
):
    # type: (...) -> BuildConfiguration
    return compute_pip_configuration(parser, args).build_configuration


def compute_repos_configuration(
    parser,  # type: ArgumentParser
    args,  # type: List[str]
):
    # type: (...) -> ReposConfiguration
    pip_configuration = compute_pip_configuration(parser, args)
    return pip_configuration.repos_configuration


def test_clp_no_pypi_option(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    repos_configuration = compute_repos_configuration(parser, args=[])
    assert len(repos_configuration.indexes) == 1
    assert len(repos_configuration.find_links) == 0

    repos_configuration = compute_repos_configuration(parser, args=["--no-pypi"])
    assert len(repos_configuration.indexes) == 0, "--no-pypi should remove the pypi index."
    assert len(repos_configuration.find_links) == 0


def test_clp_pypi_option_duplicate(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    repos_configuration = compute_repos_configuration(parser, args=[])
    assert len(repos_configuration.indexes) == 1
    assert len(repos_configuration.find_links) == 0

    repos_configuration2 = compute_repos_configuration(parser, args=["--pypi"])
    assert len(repos_configuration2.indexes) == 1
    assert len(repos_configuration2.find_links) == 0

    assert repos_configuration.indexes == repos_configuration2.indexes


def test_clp_find_links_option(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    repos_configuration = compute_repos_configuration(parser, args=["-f", "http://www.example.com"])
    assert len(repos_configuration.indexes) == 1
    assert len(repos_configuration.find_links) == 1


def test_clp_index_option(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    repos_configuration = compute_repos_configuration(parser, args=[])
    assert len(repos_configuration.indexes) == 1
    assert len(repos_configuration.find_links) == 0

    repos_configuration2 = compute_repos_configuration(
        parser, args=["-i", "http://www.example.com"]
    )
    assert len(repos_configuration2.indexes) == 2
    assert len(repos_configuration2.find_links) == 0

    assert repos_configuration2.indexes[0] == repos_configuration.indexes[0]
    assert repos_configuration2.indexes[1] == "http://www.example.com"


def test_clp_index_option_render(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    repos_configuration = compute_repos_configuration(
        parser, args=["--index", "http://www.example.com"]
    )
    assert (resolver_configuration.PYPI, "http://www.example.com") == repos_configuration.indexes
    assert () == repos_configuration.find_links


def test_clp_build_precedence(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    pip_configuration = compute_pip_configuration(parser, args=["--no-build"])
    assert not pip_configuration.build_configuration.allow_builds

    pip_configuration = compute_pip_configuration(parser, args=["--build"])
    assert pip_configuration.build_configuration.allow_builds

    pip_configuration = compute_pip_configuration(parser, args=["--no-wheel"])
    assert not pip_configuration.build_configuration.allow_wheels

    pip_configuration = compute_pip_configuration(parser, args=["--wheel"])
    assert pip_configuration.build_configuration.allow_wheels


def test_pex_repository(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser, include_pex_repository=True)

    resolver_configuration = compute_resolver_configuration(
        parser, args=["--pex-repository", "a.pex"]
    )
    assert isinstance(resolver_configuration, PexRepositoryConfiguration)
    assert "a.pex" == resolver_configuration.pex_repository


def test_invalid_configuration(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser, include_pex_repository=True)

    with pytest.raises(resolver_options.InvalidConfigurationError):
        compute_resolver_configuration(
            parser, args=["--pex-repository", "a.pex", "-f", "https://a.find/links/repo"]
        )


def test_vendored_pip_version(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    pip_configuration = compute_pip_configuration(parser, args=["--pip-version", "vendored"])
    assert pip_configuration.version is PipVersion.VENDORED

    pip_configuration = compute_pip_configuration(parser, args=["--pip-version", "20.3.4-patched"])
    assert pip_configuration.version is PipVersion.VENDORED


def test_latest_pip_version(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    pip_configuration = compute_pip_configuration(parser, args=["--pip-version", "latest"])
    assert pip_configuration.version is PipVersion.LATEST


def test_resolver_version_invalid(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.raises(resolver_options.InvalidConfigurationError):
        compute_pip_configuration(
            parser, args=["--pip-version", "23.2", "--resolver-version", "pip-legacy-resolver"]
        )


def test_build_configuration_default(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    build_configuration = compute_build_configuration(parser, args=[])
    assert PipConfiguration().build_configuration == build_configuration
    assert BuildConfiguration() == build_configuration


def test_build_configuration_invalid_no_builds_no_wheels(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.raises(
        BuildConfiguration.Error,
        match=re.escape(
            "Cannot both disallow builds and disallow wheels. Please allow one of these or both so "
            "that some distributions can be resolved."
        ),
    ):
        compute_build_configuration(parser, args=["--no-build", "--no-wheel"])


def test_build_configuration_invalid_no_builds_only_build(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.raises(
        BuildConfiguration.Error,
        match=re.escape(
            "Builds were disallowed, but the following project names are configured to only allow "
            "building: ansicolors"
        ),
    ):
        compute_build_configuration(parser, args=["--no-build", "--only-build", "ansicolors"])


def test_build_configuration_invalid_no_wheels_only_wheel(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.raises(
        BuildConfiguration.Error,
        match=re.escape(
            "Resolving wheels was disallowed, but the following project names are configured to "
            "only allow resolving pre-built wheels: ansicolors, cowsay"
        ),
    ):
        compute_build_configuration(
            parser, args=["--no-wheel", "--only-wheel", "cowsay", "--only-wheel", "ansicolors"]
        )


def test_build_configuration_invalid_only_build_only_wheel(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.raises(
        BuildConfiguration.Error,
        match=re.escape(
            "The following project names were specified as only being allowed to be built and only "
            "allowed to be resolved as pre-built wheels, please pick one or the other for each: "
            "cowsay"
        ),
    ):
        compute_build_configuration(
            parser,
            args=["--only-wheel", "cowsay", "--only-wheel", "ansicolors", "--only-build", "cowsay"],
        )


def test_build_configuration_warn_prefer_older_binary_unused(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.warns(
        PEXWarning,
        match=re.escape(
            "The prefer older binary setting was requested, but this has no effect unless both "
            "pre-built wheels and sdist builds are allowed."
        ),
    ):
        build_configuration = compute_build_configuration(
            parser, args=["--prefer-binary", "--no-build"]
        )
        assert not build_configuration.allow_builds
        assert build_configuration.prefer_older_binary

    with pytest.warns(
        PEXWarning,
        match=re.escape(
            "The prefer older binary setting was requested, but this has no effect unless both "
            "pre-built wheels and sdist builds are allowed."
        ),
    ):
        build_configuration = compute_build_configuration(
            parser, args=["--prefer-binary", "--no-wheel"]
        )
        assert not build_configuration.allow_wheels
        assert build_configuration.prefer_older_binary


def test_build_configuration_warn_use_pep517_no_build(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.warns(
        PEXWarning,
        match=re.escape(
            "Use of PEP-517 builds was set to True, but builds are turned off; so this setting has "
            "no effect."
        ),
    ):
        build_configuration = compute_build_configuration(
            parser, args=["--use-pep517", "--no-build"]
        )
        assert not build_configuration.allow_builds
        assert build_configuration.use_pep517


def test_build_configuration_warn_no_build_isolation_no_build(parser):
    # type: (ArgumentParser) -> None
    resolver_options.register(parser)

    with pytest.warns(
        PEXWarning,
        match=re.escape(
            "Build isolation was turned off, but builds are also turned off; so this setting has "
            "no effect."
        ),
    ):
        build_configuration = compute_build_configuration(
            parser, args=["--no-build-isolation", "--no-build"]
        )
        assert not build_configuration.allow_builds
        assert not build_configuration.build_isolation



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import shutil
import subprocess

from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.mitmproxy import Proxy

if TYPE_CHECKING:
    from typing import Any


def test_rel_cert_path(
    proxy,  # type: Proxy
    tmpdir,  # type: Any
):
    # type: (...) -> None
    pex_file = os.path.join(str(tmpdir), "pex")
    workdir = os.path.join(str(tmpdir), "workdir")
    os.mkdir(workdir)
    with proxy.run() as (port, ca_cert):
        shutil.copy(ca_cert, os.path.join(workdir, "cert"))
        run_pex_command(
            args=[
                "--proxy",
                "http://localhost:{port}".format(port=port),
                "--cert",
                "cert",
                # N.B.: The original issue (https://github.com/pex-tool/pex/issues/1537) involved
                # avro-python3 1.10.0, but that distribution utilizes setup_requires which leads to
                # issues in CI for Mac. We use the Python 2/3 version of the same distribution
                # instead, which had setup_requires removed in
                # https://github.com/apache/avro/pull/818.
                "avro==1.10.0",
                "-o",
                pex_file,
            ],
            cwd=workdir,
        ).assert_success()
        subprocess.check_call(args=[pex_file, "-c", "import avro"])



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os.path
import subprocess
from textwrap import dedent

import pytest

import testing
from pex.common import safe_open
from pex.interpreter_constraints import InterpreterConstraint
from pex.pep_503 import ProjectName
from pex.requirements import PyPIRequirement, parse_requirement_file
from pex.resolve.lockfile import json_codec
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InstallationChoice, Virtualenv
from testing.cli import run_pex3
from testing.lock import index_lock_artifacts

if TYPE_CHECKING:
    from typing import Any


def create_entry_module(project_dir):
    # type: (str) -> None

    with safe_open(os.path.join(project_dir, "entry.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                #!/usr/bin/env python3

                import numpy  # not used, just to illustrate a dependency


                def main():
                    print("Hello ", numpy.__version__)
                """
            )
        )


def create_setup_py(
    project_dir,  # type: str
    **setup_kwargs  # type: Any
):
    # type: (...) -> None

    with safe_open(os.path.join(project_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                #!/usr/bin/env python3

                import setuptools

                setuptools.setup(
                    name="repro",
                    version="0.0.1",
                    python_requires="~=3.9",
                    dependency_links=[],
                    entry_points={{
                        "console_scripts": [
                            "entry = entry:main",
                        ],
                    }},
                    py_modules=["entry"],
                    **{setup_kwargs!r}
                )
                """
            ).format(setup_kwargs=setup_kwargs)
        )


@pytest.fixture
def issue_2412_repro_project_dir(tmpdir):
    # type: (Any) -> str

    project_dir = os.path.join(str(tmpdir), "project")

    # This exactly replicates the project setup in the original problem repro repo in
    # https://github.com/pex-tool/pex/issues/2412 sans Makefile.
    create_entry_module(project_dir)
    input_requirements = os.path.join(project_dir, "requirements.in")
    with safe_open(input_requirements, "w") as fp:
        print("numpy", file=fp)
    create_setup_py(project_dir)
    return project_dir


def assert_bdist_pex(
    project_dir,  # type: str
    expected_numpy_version,  # type: str
):
    # type: (...) -> None

    assert "Hello  {numpy_version}\n".format(
        numpy_version=expected_numpy_version
    ) == subprocess.check_output(args=[os.path.join(project_dir, "dist", "entry")]).decode("utf-8")


skip_if_incompatible_with_repro_project = pytest.mark.skipif(
    not InterpreterConstraint.matches("CPython>=3.9,<3.13"),
    reason=(
        "The repro project under test requires Python ~=3.9 and we further restrict to CPython "
        "and place a version ceiling at 3.13 to ensure we resolve numpy wheels and do not need to "
        "build the sdist."
    ),
)


@skip_if_incompatible_with_repro_project
def test_bdist_pex_locked_issue_2412_repro_exact(
    issue_2412_repro_project_dir,  # type: str
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    venv = Virtualenv.create(
        venv_dir=os.path.join(issue_2412_repro_project_dir, ".env"),
        install_pip=InstallationChoice.UPGRADED,
        install_setuptools=InstallationChoice.UPGRADED,
    )
    subprocess.check_call(
        args=[venv.bin_path("pip"), "install", "-U", "pip-tools", pex_project_dir]
    )

    subprocess.check_call(
        args=[
            venv.bin_path("pip-compile"),
            "--output-file",
            "requirements.txt",
            "--generate-hashes",
            "requirements.in",
        ],
        cwd=issue_2412_repro_project_dir,
    )
    locked_requirements = list(
        parse_requirement_file(os.path.join(issue_2412_repro_project_dir, "requirements.txt"))
    )
    assert 1 == len(locked_requirements)

    locked_requirement = locked_requirements[0]
    assert isinstance(locked_requirement, PyPIRequirement)
    assert (
        "--hash=sha256:" in locked_requirement.line.raw_text
    ), "Expected the compiled requirements file to include hashes."

    locked_numpy_specifiers = list(locked_requirement.requirement.specifier)
    assert 1 == len(locked_numpy_specifiers)
    locked_numpy_specifier = locked_numpy_specifiers[0]
    assert "==" == locked_numpy_specifier.operator
    locked_numpy_version = locked_numpy_specifier.version

    subprocess.check_call(
        args=[
            venv.interpreter.binary,
            "setup.py",
            "bdist_pex",
            "--pex-args",
            "--disable-cache -vvvv -r requirements.txt --pip-version 24.0",
            "--bdist-all",
        ],
        cwd=issue_2412_repro_project_dir,
    )
    assert_bdist_pex(issue_2412_repro_project_dir, expected_numpy_version=locked_numpy_version)


def assert_bdist_pex_locked(
    project_dir,  # type: str
    lock,  # type: str
):
    # type: (...) -> None

    venv = Virtualenv.create(
        venv_dir=os.path.join(project_dir, ".env"),
        install_pip=InstallationChoice.YES,
        install_setuptools=InstallationChoice.YES,
    )
    subprocess.check_call(args=[venv.bin_path("pip"), "install", testing.pex_project_dir()])
    subprocess.check_call(
        args=[
            venv.interpreter.binary,
            "setup.py",
            "bdist_pex",
            "--pex-args",
            "--disable-cache -vvvv --lock {lock}".format(lock=lock),
            "--bdist-all",
        ],
        cwd=project_dir,
    )

    locked_numpy_version = index_lock_artifacts(json_codec.load(lock))[
        ProjectName("numpy")
    ].pin.version.raw
    assert_bdist_pex(project_dir=project_dir, expected_numpy_version=locked_numpy_version)


@skip_if_incompatible_with_repro_project
def test_bdist_pex_locked_issue_2412_repro_pex_lock(
    issue_2412_repro_project_dir,  # type: str
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    lock = os.path.join(issue_2412_repro_project_dir, "lock.json")
    run_pex3(
        "lock",
        "sync",
        "-r",
        os.path.join(issue_2412_repro_project_dir, "requirements.in"),
        "--pip-version",
        "24.0",
        "--style",
        "universal",
        "--interpreter-constraint",
        "~=3.9",
        "--indent",
        "2",
        "--lock",
        lock,
    ).assert_success()
    assert_bdist_pex_locked(project_dir=issue_2412_repro_project_dir, lock=lock)


@skip_if_incompatible_with_repro_project
def test_bdist_pex_locked_issue_2412_repro_pex_lock_inlined_requirements(
    tmpdir,  # type: str
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    project_dir = os.path.join(str(tmpdir), "project")
    create_entry_module(project_dir)
    create_setup_py(project_dir, install_requires=["numpy"])

    lock = os.path.join(project_dir, "lock.json")
    run_pex3(
        "lock",
        "sync",
        "--project",
        project_dir,
        "--pip-version",
        "24.0",
        "--style",
        "universal",
        "--interpreter-constraint",
        "~=3.9",
        "--indent",
        "2",
        "--lock",
        lock,
    ).assert_success()
    assert_bdist_pex_locked(project_dir=project_dir, lock=lock)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path

from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def root(doc_type="html"):
    # type: (str) -> Optional[str]

    doc_root = os.path.join(os.path.dirname(__file__), doc_type)
    return doc_root if os.path.isdir(doc_root) else None



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.interpreter_constraints import InterpreterConstraint
from pex.pep_503 import ProjectName
from pex.resolve.locked_resolve import LockedRequirement
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Mapping, Union


def index_lock_artifacts(lock_file):
    # type: (Union[str, Lockfile]) -> Mapping[ProjectName, LockedRequirement]

    lock = lock_file if isinstance(lock_file, Lockfile) else json_codec.load(lock_file)
    assert 1 == len(lock.locked_resolves)
    locked_resolve = lock.locked_resolves[0]
    return {
        locked_requirement.pin.project_name: locked_requirement
        for locked_requirement in locked_resolve.locked_requirements
    }


def extract_lock_option_args(lock_file):
    # type: (Union[str, Lockfile]) -> List[str]

    lock = lock_file if isinstance(lock_file, Lockfile) else json_codec.load(lock_file)
    lock_args = [
        "--pip-version",
        str(lock.pip_version),
        "--resolver-version",
        str(lock.resolver_version),
        "--style",
        str(lock.style),
    ]
    for requires_python in lock.requires_python:
        lock_args.append("--interpreter-constraint")
        lock_args.append(str(InterpreterConstraint.parse(requires_python)))
    for target_system in lock.target_systems:
        lock_args.append("--target-system")
        lock_args.append(str(target_system))
    return lock_args
